/*
  MCU Firmware for Peristaltic Pump Motor Control

  Copyright (c) 20015 Frank Z. Zhao

  This file is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This file is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  First Revision 7 July 2015 by Frank Zhao
  Designed for ATmega168 or similar
*/

#ifndef F_CPU
#define F_CPU 8000000
#endif

#include <avr/io.h>
#include <util/delay.h>

#define BTN_FWD_PORTx PORTD
#define BTN_REV_PORTx PORTD
#define BTN_CTR_PORTx PORTD
#define BTN_FWD_DDRx DDRD
#define BTN_REV_DDRx DDRD
#define BTN_CTR_DDRx DDRD
#define BTN_FWD_PINx PIND
#define BTN_REV_PINx PIND
#define BTN_CTR_PINx PIND
#define BTN_FWD_BP 2
#define BTN_REV_BP 4
#define BTN_CTR_BP 3

#define LED_DDRx DDRD
#define LED_PORTx PORTD
#define LED_BP 7

#define ADCCHAN_POT_SPEED 0
#define ADCCHAN_POT_INTERVAL 1

#define PWM_OCR OCR0B
#define PWM_COM _BV(COM0B1)
#define DIR_PORTx PORTD
#define DIR_DDRx DDRD
#define DIR_BP 6

uint16_t adc_read(uint8_t chan)
{
	static char hasinit = 0;
	volatile int16_t res;
	if (hasinit == 0)
	{
		ADCSRA = _BV(ADEN) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);
		hasinit = 1;
	}
	DDRC &= ~_BV(chan);
	DIDR0 |= (1 << chan); // disables digital input on that channel to save power and get slightly better reading (not sure about this)
	ADMUX = _BV(REFS0) | chan; // use AVCC as reference
	ADCSRA |= _BV(ADSC); // start
	loop_until_bit_is_clear(ADCSRA, ADSC);

	// must be read in correct order to prevent locking the register
	res = ADCL;
	res += ADCH << 8;

	return res;
}

void timer_start()
{
	PORTD &= ~(_BV(5) | _BV(6)); // pins off
	DDRD |= _BV(5) | _BV(6); // pins as output

	// enables OCR0B pin in fast PWM mode
	TCCR0A = PWM_COM | _BV(WGM01) | _BV(WGM00);

	// both off
	OCR0A = 0;
	OCR0B = 0;

	TCCR0B = _BV(CS00); // start
}

int main()
{
	char latched = 0; // sticky slow interval mode status flag
	char led = 0; // LED is turned ON or OFF at the end of the main loop
	uint32_t intervalcnt = 0;

	// setup inputs, with internal pull-up resistors
	BTN_REV_DDRx &= ~_BV(BTN_REV_BP);
	BTN_FWD_DDRx &= ~_BV(BTN_FWD_BP);
	BTN_CTR_DDRx &= ~_BV(BTN_CTR_BP);
	BTN_REV_PORTx |= _BV(BTN_REV_BP);
	BTN_FWD_PORTx |= _BV(BTN_FWD_BP);
	BTN_CTR_PORTx |= _BV(BTN_CTR_BP);
	
	// start motor off
	DIR_PORTx &= ~_BV(DIR_BP);
	DIR_DDRx |= _BV(DIR_BP);

	timer_start();

	while (1) // main program loop
	{
		led = 0;

		if (bit_is_clear(BTN_REV_PINx, BTN_REV_BP))
		{
			// full reverse speed
			DIR_PORTx &= ~_BV(DIR_BP);
			PWM_OCR = 0xFF;
			latched = 0; // cancel latched mode
			led = 1;
			_delay_ms(100); // debounce
		}
		else if (bit_is_clear(BTN_FWD_PINx, BTN_FWD_BP))
		{
			// full forward speed
			DIR_PORTx |= _BV(DIR_BP);
			PWM_OCR = 0xFF;
			latched = 0; // cancel latched mode
			led = 1;
			_delay_ms(100); // debounce
		}
		else if (bit_is_clear(BTN_CTR_PINx, BTN_CTR_BP))
		{
			if (latched == 0) // initial press
			{
				latched = 1;
				intervalcnt = 0; // restart interval timer
			}
			else if (latched == 2) // unlatching press
			{
				latched = 3;
			}
			_delay_ms(100); // debounce
		}
		else // nothing pressed
		{
			if (latched == 3) // recent unlatch
			{
				latched = 0;
				_delay_ms(100); // debounce
			}
			if (latched == 0)
			{
				// motors off
				PWM_OCR = 0;
				DIR_PORTx &= ~_BV(DIR_BP);
			}
		}

		if (latched == 1) // was initial press
		{
			if (bit_is_set(BTN_CTR_PINx, BTN_CTR_BP)) // check for release
			{
				_delay_ms(100); // debounce
				latched = 2;
			}
		}

		if (latched == 1 || latched == 2)
		{
			led = 1; // LED on even if not spinning
			intervalcnt += 1;
			int16_t r = adc_read(ADCCHAN_POT_INTERVAL);
			r -= 8; // for potentiometer offset. bottom of potentiometer should equate to no off period at all
			if (r <= 0)
			{
				r = 0;
				intervalcnt = 0; // keep on
			}
			if (intervalcnt <= 222) // about 2 seconds
			{
				int16_t x16 = adc_read(ADCCHAN_POT_SPEED);

				x16 >>= 2; // rescale
				x16 += 32; // this gives the result a minimum speed, and also ensures the top of the pot reading is at least 255

				x16 = (x16 > 255) ? 255 : ((x16 < 0) ? 0 : x16); // limit to valid values

				// set speed
				DIR_PORTx |= _BV(DIR_BP);
				PWM_OCR = (uint8_t)x16;
			}
			else
			{
				// motor off
				PWM_OCR = 0;
				DIR_PORTx &= ~_BV(DIR_BP);
				if (r <= intervalcnt)
				{
					intervalcnt = 0; // roll-over
				}
			}
			_delay_ms(9); // the top of the potentiometer should equate to around 10 seconds
		}

		if (led != 0)
		{
			LED_DDRx |= _BV(LED_BP); // pin as output
			LED_PORTx |= _BV(LED_BP); // light
		}
		else
		{
			LED_PORTx &= ~_BV(LED_BP); // off
		}
	}
	return 0;
}